
/* 
 * Take co-registered LULC and DEM data in *binary* GRASS format, 
 * read in a specified subsection of that data, and write it to files
 * as both a binary PPM file (output_file_name.ppm) and an ASCII file 
 * containing elevation and color data (output_file_name.data).
 *
 * Program is called with input variables 
 *   a.out <DEM data input file> <LULC data input file> <output_file_name>
 * and reads in LULC color map from file LULC.colors,
 * and the coordinates of the area from stdin
 * 
*/


#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#define NumCategories 100
#define NumColors     3

#define GridSpacing 100

#define LineSize 4096

char color[NumCategories][NumColors];  /* color map for LULC categories */

/* Coordinates of subset of data to be output */
int x0, x1, y0, y1;

main(int argc, char *argv[])
{
   FILE  *ColorMap;
   FILE  *OutputData;
   int    InputDEM;
   int    InputLULC;
   int    OutputPPM;
   int    BogusDataFile;
   char   Header[256];
   char   filename[256];
   char   line[LineSize];
   int    North = 0;
   int    South = 0;
   int    East = 0;
   int    West = 0;
   int    M = 0;
   int    N = 0;
   short *DEM_Buffer;
   short *Bogus_DEM_Buffer;
   char  *LULC_Buffer;
   char  *Bogus_LULC_Buffer;
   char  *Color_Buffer;
   int    DEM_Buffer_Size;
   int    LULC_Buffer_Size;
   int    Color_Buffer_Size;
   int    i,j, tmp;
   int    Index, category;
   short  red, green, blue;
   short  DEM;
   short  LULC;
   
   if (argc < 3)
   {
      printf("usage: <GRASS DEM datafile> <GRASS LULC datafile> <output \
file name>\n");
      exit(1);
   }
   
   /* Open input file of binary DEM data */
   sprintf(filename, "%s.BINARY", argv[1]);
   if ((InputDEM = open(filename, O_RDONLY)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   /* Read in the coordinates for the binary DEM data */
   read(InputDEM, &North, sizeof(int));
   read(InputDEM, &South, sizeof(int));
   read(InputDEM, &East,  sizeof(int));
   read(InputDEM, &West,  sizeof(int));
   read(InputDEM, &M, sizeof(int));
   read(InputDEM, &N, sizeof(int));

   printf("The data covers an area of %d x %d grid points ", M, N);
   printf("with a grid spacing of %d meters.\n\n", GridSpacing);
   printf("The area has UTM coordinates (in meters) of \n"); 
   printf(" (North, South, East, West) = (%d, %d, %d, %d)\n\n",
           North, South, East, West);

   /* Read in the coordinates for the binary LULC data */
   /*  (and check that they are the same as for the DEM data?) */
   sprintf(filename, "%s.BINARY", argv[2]);
   if ((InputLULC = open(filename, O_RDONLY)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   read(InputLULC, &North, sizeof(int));
   read(InputLULC, &South, sizeof(int));
   read(InputLULC, &East,  sizeof(int));
   read(InputLULC, &West,  sizeof(int));
   read(InputLULC, &M, sizeof(int));
   read(InputLULC, &N, sizeof(int));

   /* Read in the requested data coordinates */

   printf("Enter the (N, S, E, W) co-ordinates of the section of this area\n");
   printf("you wish to display, in units of grid points\n");
   printf("(0 <= S <= N <= %ld, 0 <= W <= E <= %ld) \n", M, N);

   printf("\n South? "); scanf("%d", &x0);
   printf("\n North? "); scanf("%d", &x1);
   printf("\n West?  "); scanf("%d", &y0);
   printf("\n East?  "); scanf("%d", &y1);

   /* Fix up bogus data requests */

   if ((x0 > M) || (x1 > M)) {
      printf("North - South coordinates must be less than %ld\n", M);
      fflush(stdout);
      exit(4);
   }
   if ((y0 > N) || (y1 > N)) {
      printf("East - West coordinates must be less than %ld\n", N);
      fflush(stdout);
      exit(5);
   }
   if (x0 > x1) {
     tmp = x0; x0 = x1; x1 = tmp;
      printf("Changed input values so that North > South\n");
   }
   if (y0 > y1) {
     tmp = y0; y0 = y1; y1 = tmp;
      printf("Changed input values so that East > West\n");
   }

   printf("\nSouth = %d  North = %d  West = %d  East = %d\n\n", x0, x1, y0, y1);
   
   
   /* Open output file for DEM and color data */
   sprintf(filename, "%s.data", argv[3]);
   if ((OutputData = fopen(filename, "w")) == NULL)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   fprintf(OutputData, "north: %d\n", South + x1*GridSpacing); 
   fprintf(OutputData, "south: %d\n", South + x0*GridSpacing); 
   fprintf(OutputData, "east: %d\n", West + y1*GridSpacing); 
   fprintf(OutputData, "west: %d\n", West + y0*GridSpacing); 
   fprintf(OutputData, "rows: %d\n", x1-x0);
   fprintf(OutputData, "cols: %d\n", y1-y0);


   /* Set up data arrays */

   DEM_Buffer_Size = N * sizeof(short);
   LULC_Buffer_Size = N * sizeof(char);
   Color_Buffer_Size = 3 * N * sizeof(char);

   if ((DEM_Buffer = (short *) malloc(DEM_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }

   if ((LULC_Buffer = (char *) malloc(LULC_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }
   if ((Color_Buffer = (char *) malloc(Color_Buffer_Size)) == NULL)
   {
      printf("Out of memory!\n");
      exit(1);
   }


   /* Open output PPM image file for processed data */
   sprintf(filename, "%s.ppm", argv[3]);
   if ((OutputPPM = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
                          S_IRUSR |  S_IWUSR | S_IRGRP)) == -1)
   {
      printf("Error -- cannot open %s\n", filename);
      exit(1);
   }
   sprintf(Header, "P6\n%d %d\n255\n\0", y1-y0,  x1-x0);
   write(OutputPPM, Header, strlen(Header));


   for (i = 0; i < NumCategories; i++)
      for (j = 0; j < NumColors; j++)
	 color[i][j] = (char) 0;


   /* Read in the LULC color map */
   printf("Reading in the LULC color map ....\n");
   fflush(stdout);
   if ((ColorMap = fopen("LULC.color.map", "r")) == NULL)
   {
      printf("Error -- cannot open file LULC.color\n");
      exit(1);
   }

   while (fgets(line, LineSize, ColorMap) != NULL) 
   {
      /* screen for lines commented with a # */
      if (strncmp("#", line, 1)) 
      {
	 sscanf(line, "%d %hd %hd %hd", &category, &red, &green, &blue);
	 color[category][0] = (char) red;
 	 color[category][1] = (char) green;
	 color[category][2] = (char) blue;
	 /*
	 printf("%d  %d %d %d\n", category, 
	                          (int)color[category][0], 
				  (int)color[category][1],
				  (int)color[category][2]);
	 fflush(stdout);
	 */
      }
   }

   /* Change Color_Buffer_Size to correct value for requested area */
   Color_Buffer_Size = 3 * (y1-y0) * sizeof(char);


   /*  Read in and process the DEM and LULC data */
   /*  N.B. Data is ordered from top left to bottom right, so need to  *
    *  count down from the largest value of i (i.e. North to South)    */

   printf("Reading in and processing the DEM and LULC data ....\n");
   fflush(stdout);

   for (i = M-1; i >= x0; i--)
   {
      /* Read in rows of binary LULC and DEM data */
      read(InputDEM, DEM_Buffer, DEM_Buffer_Size);
      read(InputLULC, LULC_Buffer, LULC_Buffer_Size);
      
      /* Only process data if it is in the requested area */
      if (i < x1) {
      
        Index = 0;

        for (j = y0; j < y1; j++)
        {
           DEM = DEM_Buffer[j];
           LULC = LULC_Buffer[j];

	   Color_Buffer[Index]     = color[LULC][0];
           Color_Buffer[Index + 1] = color[LULC][1];
           Color_Buffer[Index + 2] = color[LULC][2];

	   fprintf(OutputData, "%d %d %d %d\n", DEM, 
	                       (int)color[LULC][0], 
			       (int)color[LULC][1],
			       (int)color[LULC][2]);

	   Index += 3;
        }

        /* Print out row of LULC color PPM file */
        write(OutputPPM, Color_Buffer, Color_Buffer_Size);

      }
   }

   close(InputDEM);
   close(InputLULC);
   close(OutputPPM);
   fclose(OutputData);
}

